"""
Hyfi configuration file.
"""
import os
from typing import Any, List, Optional, Union

from omegaconf import DictConfig
from pydantic import (
    BaseModel,
    ConfigDict,
    FieldValidationInfo,
    PrivateAttr,
    field_validator,
    model_validator,
)

from hyfi.about import AboutConfig
from hyfi.core import __global_hyfi__
from hyfi.dotenv import DotEnvConfig
from hyfi.pipeline import PipelineConfig
from hyfi.project import ProjectConfig
from hyfi.task import TaskConfig
from hyfi.utils.envs import ENVs
from hyfi.utils.logging import LOGGING
from hyfi.utils.notebooks import NBs
from hyfi.workflow import WorkflowConfig

logger = LOGGING.getLogger(__name__)


class HyfiConfig(BaseModel):
    """HyFI root config class.  This class is used to store the configuration"""

    debug_mode: bool = False
    resolve: bool = False
    verbose: bool = False
    logging_level: str = "WARNING"

    hydra: Optional[DictConfig] = None

    about: Optional[AboutConfig] = None
    copier: Optional[DictConfig] = None
    project: Optional[ProjectConfig] = None
    pipeline: Optional[PipelineConfig] = None
    task: Optional[TaskConfig] = None
    workflow: Optional[WorkflowConfig] = None
    tasks: Optional[List[str]] = None

    _version_: str = PrivateAttr(__global_hyfi__.version)
    _initilized_: bool = PrivateAttr(False)

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        validate_assignment=True,
        extra="allow",
    )  # type: ignore

    @model_validator(mode="before")
    def _validate_model_data(cls, data):
        """
        Validate and set the model data.

        Args:
            cls: Class to use for config lookup
            data: Dictionary of values to check and set

        Returns:
            Validated dictionary of values
        """
        key = "hyfi_config_path"
        val = ENVs.check_and_set_osenv_var(key, data.get(key))
        # Set the hyfi_config_module value in the configuration file.
        if val is not None:
            data[key] = val
            key = "hyfi_config_module"
            data[key] = ENVs.check_and_set_osenv_var(key, val.replace("pkg://", ""))
        return data

    @field_validator("logging_level")
    def _validate_logging_level(cls, v, info: FieldValidationInfo):
        """
        Validate and set the logging level
        """
        verbose = info.data.get("verbose", False)
        # Set verbose to INFO.
        if verbose and v == "WARNING":
            v = "INFO"
        logger.setLevel(v)
        return v

    def __init__(self, **config_kwargs: Any):
        """
        Initialize the object with data

        Args:
            config_kwargs: Dictionary of values to initialize the object with
        """
        super().__init__(**config_kwargs)

    def init_project(
        self,
        project_name: str = "",
        project_description: str = "",
        project_root: str = "",
        project_workspace_name: str = "",
        global_hyfi_root: str = "",
        global_workspace_name: str = "",
        num_workers: int = -1,
        log_level: str = "",
        reinit: bool = True,
        autotime: bool = True,
        retina: bool = True,
        verbose: Union[bool, int] = False,
        **kwargs,
    ):
        """
        Initialize and start hyfi.

        Args:
            project_name: Name of the project to use.
            project_description: Description of the project that will be used.
            project_root: Root directory of the project.
            project_workspace_name: Name of the project's workspace directory.
            global_hyfi_root: Root directory of the global hyfi.
            global_workspace_name: Name of the global hierachical workspace directory.
            num_workers: Number of workers to run.
            log_level: Log level for the log.
            autotime: Whether to automatically set time and / or keep track of run times.
            retina: Whether to use retina or not.
            verbose: Enables or disables logging
        """
        envs = DotEnvConfig(HYFI_VERBOSE=verbose)  # type: ignore
        # Set the project name environment variable HYFI_PROJECT_NAME environment variable if project_name is not set.
        if project_name:
            envs.HYFI_PROJECT_NAME = ENVs.expand_posix_vars(project_name)
        # Set the project description environment variable HYFI_PROJECT_DESC environment variable.
        if project_description:
            envs.HYFI_PROJECT_DESC = ENVs.expand_posix_vars(project_description)
        # Set environment variables HYFI_PROJECT_ROOT to the project root if project_root is set to true.
        if project_root:
            envs.HYFI_PROJECT_ROOT = ENVs.expand_posix_vars(project_root)
        # Set the project workspace name environment variable HYFI_PROJECT_WORKSPACE_NAME environment variable if project_workspace_name is set to the project workspace name.
        if project_workspace_name:
            envs.HYFI_PROJECT_WORKSPACE_NAME = ENVs.expand_posix_vars(
                project_workspace_name
            )
        # Expand the hyfi_root environment variable.
        if global_hyfi_root:
            envs.HYFI_GLOBAL_ROOT = ENVs.expand_posix_vars(global_hyfi_root)
        # Set the global workspace name environment variable HYFI_GLOBAL_WORKSPACE_NAME environment variable.
        if global_workspace_name:
            envs.HYFI_GLOBAL_WORKSPACE_NAME = ENVs.expand_posix_vars(
                global_workspace_name
            )
        # Set the number of workers to use.
        if num_workers:
            envs.HYFI_NUM_WORKERS = num_workers
        # Set the log level to the given log level.
        if log_level:
            envs.HYFI_LOG_LEVEL = log_level
            LOGGING.setLogger(log_level)
            logger.setLevel(log_level)
        # Load the extentions for the autotime extension.
        if autotime:
            NBs.load_extentions(exts=["autotime"])
        # Set the retina matplotlib formats.
        if retina:
            NBs.set_matplotlib_formats("retina")

        self.initialize(force=reinit)
        self.project = ProjectConfig()
        logger.info("HyFi project initialized with %s", self.project.project_name)

    def initialize(self, force: bool = False) -> None:
        """
        Initialize hyfi.

        Returns:
            A boolean indicating whether initialization was successful
        """
        # Returns the current value of the _initilized_ attribute.
        if self._initilized_ and not force:
            return
        if self.about is None:
            self.about = AboutConfig()
        ENVs.load_dotenv()

        self._initilized_ = True

    def terminate(self) -> None:
        """
        Terminate hyfi config by stopping joblib

        Returns:
            True if successful False
        """
        # If the module is not initialized yet.
        if not self._initilized_:
            return
        # Stop the backend if the joblib is running.
        if self.project and self.project.joblib:
            self.project.joblib.stop_backend()
        self._initilized_ = False

    def __repr__(self):
        """
        Returns a string representation of HyFIConfig.

        Returns:
            The string representation of HyFI
        """
        return f"HyFIConfig(project={self.project})"

    def __str__(self):
        """
        Returns a string representation of the object.

        Returns:
                The string representation of the
        """
        return self.__repr__()

    @property
    def app_version(self):
        """
        Get the version of the application.


        Returns:
            The version of the application
        """
        return __global_hyfi__.version

    @property
    def app_name(self):
        """
        Get the name of the application.

        Returns:
            The name of the application
        """
        return self.about.name if self.about else __global_hyfi__.hyfi_name

    @property
    def package_name(self):
        """
        Get the name of the package.

        Returns:
            The name of the package
        """
        return __global_hyfi__.package_name

    @property
    def dotenv(self):
        return DotEnvConfig()  # type: ignore

    @property
    def osenv(self):
        return os.environ

    def print_about(self, **kwargs):
        self.about = AboutConfig(**kwargs)
        pkg_name = self.package_name
        name = self.app_name
        print()
        for k, v in self.about.model_dump().items():
            if k.startswith("_") or k == "version":
                continue
            print(f"{k:11} : {v}")
        print(f"{'version':11} : {self.app_version}")
        if pkg_name:
            print(f"\nExecute `{pkg_name} --help` to see what you can do with {name}")

    def get_project_path(self, path_name: str) -> str:
        if self.project and self.project.path:
            return str(self.project.path.get_path(path_name))
        return ""

    @property
    def project_root_path(self):
        """Get the project root directory."""
        return str(self.project.root_dir) if self.project else "."

    @property
    def project_workspace_path(self):
        """Get the project workspace directory."""
        return str(self.project.workspace_dir) if self.project else "."


__global_config__ = HyfiConfig()


def __project_root_path__():
    """Global HyFI config path for the project root."""
    return __global_config__.project_root_path


def __project_workspace_path__():
    """Global HyFI config path for the project workspace directory."""
    return __global_config__.project_workspace_path



"""
    Configuration for HyFi Pipelines
"""
from typing import Any, Callable, Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, model_validator

from hyfi.composer import Composer
from hyfi.task import TaskConfig
from hyfi.utils.envs import ENVs
from hyfi.utils.logging import LOGGING

logger = LOGGING.getLogger(__name__)


class BaseRunConfig(BaseModel):
    """Run Configuration"""

    run_with: Optional[Dict[str, Any]] = {}

    name: Optional[str] = ""
    desc: Optional[str] = ""
    env: Optional[Dict[str, Any]] = {}
    verbose: bool = False

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="allow",
        validate_assignment=False,
        alias_generator=Composer.generate_alias_for_special_keys,
    )  # type: ignore

    @model_validator(mode="before")
    def validate_model_config_before(cls, data):
        # logger.debug("Validating model config before validating each field.")
        return Composer.replace_special_keys(Composer.to_dict(data))

    @property
    def kwargs(self):
        return self.run_with or {}


class RunningConfig(BaseRunConfig):
    """Running Configuration"""

    uses: str = ""


Steps = List[RunningConfig]
Pipelines = List[RunningConfig]
Tasks = List[RunningConfig]


class PipeConfig(BaseRunConfig):
    """Pipe Configuration"""

    pipe_target: str = ""
    run: str = ""

    use_pipe_obj: bool = True
    pipe_obj_arg_name: Optional[str] = ""
    return_pipe_obj: bool = False
    task: Optional[TaskConfig] = None

    def set_enviroment(self):
        if self.env:
            ENVs.check_and_set_osenv_vars(self.env)

    def get_pipe_func(self) -> Optional[Callable]:
        if self.pipe_target.startswith("lambda"):
            return eval(self.pipe_target)
        elif self.pipe_target:
            return Composer.partial(self.pipe_target)
        else:
            return None

    def get_run_func(self) -> Optional[Callable]:
        if self.run.startswith("lambda"):
            logger.info("Returning lambda function: %s", self.run)
            return eval(self.run)
        elif self.run:
            kwargs = self.run_with or {}
            if self.pipe_obj_arg_name:
                kwargs.pop(self.pipe_obj_arg_name)
            logger.info(
                "Returning partial function: %s with kwargs: %s", self.run, kwargs
            )
            return Composer.partial(self.run, **kwargs)
        else:
            logger.warning("No function found for %s", self)
            return None


class DataframePipeConfig(PipeConfig):
    columns_to_apply: Optional[Union[str, List[str]]] = []
    use_batcher: bool = True
    num_workers: int = 1

    @model_validator(mode="before")
    def _check_and_set_values(cls, data):
        num_workers = data.get("num_workers")
        if num_workers and num_workers > 1:
            data["use_batcher"] = True
        return data

    @property
    def columns(self) -> List[str]:
        if not self.columns_to_apply:
            return []
        return (
            [self.columns_to_apply]
            if isinstance(self.columns_to_apply, str)
            else self.columns_to_apply
        )


Pipes = List[PipeConfig]



"""Logging utilities"""
import logging
import os
import sys
import warnings


class LOGGING:
    @staticmethod
    def setLogger(
        level=None,
        force=True,
        filterwarnings_action="ignore",
        **kwargs,
    ) -> None:
        """
        Set the logging level and format. This is a wrapper around logging. basicConfig to allow the user to specify a different logging level for each test and to filter warnings that are not caught by the filterwarnings_action

        Args:
            level: The logging level to use
            force: If True ( default ) all warnings will be logged even if there are no warnings
            filterwarnings_action: The action to call when a warning is logged

        Returns:
            The logger that was
        """
        level = level or os.environ.get("HYFI_LOG_LEVEL") or "INFO"
        level = level.upper()
        os.environ["HYFI_LOG_LEVEL"] = level
        # Filter warnings by applying filterwarnings_action to the warnings.
        if filterwarnings_action is not None:
            warnings.filterwarnings(filterwarnings_action)  # type: ignore
        # Return the logging level.
        if isinstance(level, str):
            level = getattr(logging, level.upper(), logging.INFO)
        # Configure logging level level and force logging.
        if sys.version_info >= (3, 8):
            logging.basicConfig(level=level, force=force, **kwargs)
        else:
            logging.basicConfig(level=level, **kwargs)

    @staticmethod
    def getLogger(
        _name=None,
        _log_level=None,
        **kwargs,
    ) -> logging.Logger:
        """
        Get a logger with a given name and log level. It is possible to pass a logger name and log level to this function.

        Args:
            _name: The name of the logger to get. If not specified the name of the module is used.
            _log_level: The log level to set.

        Returns:
            The logger with the given name and log level set to the value specified in HYFI_LOG_LEVEL
        """
        _name = _name or __name__
        logger = logging.getLogger(_name)
        _log_level = _log_level or os.environ.get("HYFI_LOG_LEVEL") or "INFO"
        logger.setLevel(_log_level)
        return logger



import logging
import os
import shutil
import tempfile
from pathlib import Path

from .common import get_cache_dir, set_cache_dir


class BaseTestClass:
    """
    A custom testing class that disables some of the more verbose
    logging and that creates and destroys a temp directory as a test fixture.
    """

    PROJECT_ROOT = (Path(__file__).parent / "../../../").resolve()
    MODULE_ROOT = PROJECT_ROOT / "src/hyfi/cached_path"
    TESTS_ROOT = PROJECT_ROOT / "tests"
    FIXTURES_ROOT = PROJECT_ROOT / "tests/fixtures"

    def setup_method(self):
        logging.basicConfig(
            format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
            level=logging.DEBUG,
        )
        # Disabling some of the more verbose logging statements that typically aren't very helpful
        # in tests.
        logging.getLogger("urllib3.connectionpool").disabled = True

        self.TEST_DIR = Path(tempfile.mkdtemp(prefix="cached_path_tests"))
        os.makedirs(self.TEST_DIR, exist_ok=True)

        self._initial_cache_dir = get_cache_dir()
        set_cache_dir(self.TEST_DIR)

    def teardown_method(self):
        set_cache_dir(self._initial_cache_dir)
        shutil.rmtree(self.TEST_DIR)



"""Image utils."""
import io
import os
import platform
from pathlib import Path
from typing import List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import font_manager, rc
from PIL import Image, ImageFont

from hyfi.utils.iolibs import IOLIBs
from hyfi.utils.logging import LOGGING

logger = LOGGING.getLogger(__name__)


def scale_image(
    image: Image.Image,
    max_width: Optional[int] = 0,
    max_height: Optional[int] = 0,
    max_pixels: Optional[int] = 0,
    scale: Optional[float] = 1.0,
    resize_to_multiple_of: Optional[int] = 0,
    resample: int = Image.LANCZOS,
) -> Image.Image:
    """Scale image to have at most `max_pixels` pixels."""
    if resize_to_multiple_of is None:
        resize_to_multiple_of = 0

    w, h = image.size

    max_width = max_width or 0
    max_height = max_height or 0
    max_pixels = max_pixels or 0
    if max_width <= 0 and max_height > 0:
        max_width = int(w * max_height / h)
    elif max_height <= 0 and max_width > 0:
        max_height = int(h * max_width / w)

    if max_width > 0 and max_height > 0:
        max_pixels = max_width * max_height

    scale = np.sqrt(max_pixels / (w * h)) if max_pixels > 0 else scale or 1.0
    max_width = int(w * scale)
    max_height = int(h * scale)
    if resize_to_multiple_of > 0:
        max_width = (max_width // resize_to_multiple_of) * resize_to_multiple_of
        max_height = (max_height // resize_to_multiple_of) * resize_to_multiple_of

    if scale < 1.0 or w > max_width or h > max_height:
        image = image.resize((max_width, max_height), resample=resample)  # type: ignore
    return image


def load_image(
    image_or_uri: Union[str, Image.Image],
    max_width: Optional[int] = 0,
    max_height: Optional[int] = 0,
    max_pixels: Optional[int] = 0,
    scale: Optional[float] = 1.0,
    resize_to_multiple_of: Optional[int] = 0,
    crop_box=None,
    mode="RGB",
    **kwargs,
) -> Image.Image:
    """Load image from file or URI."""
    from PIL import Image

    if max_width is None:
        max_width = 0
    if max_height is None:
        max_height = 0
    if max_pixels is None:
        max_pixels = 0
    if isinstance(image_or_uri, Image.Image):
        img = image_or_uri.convert(mode)
    elif Path(image_or_uri).is_file():
        img = Image.open(image_or_uri).convert(mode)
    else:
        img = Image.open(io.BytesIO(IOLIBs.read(image_or_uri, **kwargs))).convert(mode)
    img = scale_image(
        img,
        max_width=max_width,
        max_height=max_height,
        max_pixels=max_pixels,
        scale=scale,
        resize_to_multiple_of=resize_to_multiple_of,
    )
    if crop_box is not None:
        img = img.crop(crop_box)
    return img


def load_images(
    images_or_uris: List[Union[str, Image.Image]],
    max_width: Optional[int] = 0,
    max_height: Optional[int] = 0,
    max_pixels: Optional[int] = 0,
    scale: Optional[float] = 1.0,
    resize_to_multiple_of: Optional[int] = 0,
    crop_to_min_size: bool = False,
    mode: str = "RGB",
    **kwargs,
) -> List[Image.Image]:
    """Load images from files or URIs."""
    imgs = [
        load_image(
            image_or_uri,
            max_width=max_width,
            max_height=max_height,
            max_pixels=max_pixels,
            scale=scale,
            resize_to_multiple_of=resize_to_multiple_of,
            mode=mode,
            **kwargs,
        )
        for image_or_uri in images_or_uris
    ]
    if crop_to_min_size:
        min_width = min(img.width for img in imgs)
        min_height = min(img.height for img in imgs)
        if resize_to_multiple_of is not None:
            min_width = (min_width // resize_to_multiple_of) * resize_to_multiple_of
            min_height = (min_height // resize_to_multiple_of) * resize_to_multiple_of
        imgs = [img.crop((0, 0, min_width, min_height)) for img in imgs]

    return imgs


def get_image_font(
    fontname: Optional[str] = None,
    fontsize: int = 12,
    lang: str = "en",
):
    """Get font for PIL image."""
    fontname, fontpath = get_plot_font(set_font_for_matplot=False, fontname=fontname)
    return ImageFont.truetype(fontpath, fontsize) if fontpath else None


def get_default_system_font(
    fontname: Optional[str] = None,
    fontpath: Optional[str] = None,
    lang: str = "ko",
    verbose: bool = False,
):
    if platform.system() == "Darwin":
        default_fontname = "AppleGothic.ttf" if lang == "ko" else "Arial.ttf"
        fontname = fontname or default_fontname
        fontpath = os.path.join("/System/Library/Fonts/Supplemental/", fontname)
    elif platform.system() == "Windows":
        default_fontname = "malgun.ttf" if lang == "ko" else "arial.ttf"
        fontname = fontname or default_fontname
        fontpath = os.path.join("c:/Windows/Fonts/", fontname)
    elif platform.system() == "Linux":
        default_fontname = "NanumGothic.ttf" if lang == "ko" else "DejaVuSans.ttf"
        fontname = fontname or default_fontname
        if fontname.lower().startswith("nanum"):
            fontpath = os.path.join("/usr/share/fonts/truetype/nanum/", fontname)
        else:
            fontpath = os.path.join("/usr/share/fonts/truetype/", fontname)
    if fontpath and not Path(fontpath).is_file():
        paths = find_font_file(fontname)
        fontpath = paths[0] if len(paths) > 0 else ""
    if verbose:
        logger.info(f"Font path: {fontpath}")
    return fontname, fontpath


def get_plot_font(
    set_font_for_matplot: bool = True,
    fontpath: Optional[str] = None,
    fontname: Optional[str] = None,
    lang: str = "en",
    verbose: bool = False,
):
    """Get font for plot"""
    if fontname and not fontname.endswith(".ttf"):
        fontname += ".ttf"
    if not fontpath:
        fontname, fontpath = get_default_system_font(fontname, fontpath, lang, verbose)

    if fontpath and Path(fontpath).is_file():
        font_manager.fontManager.addfont(fontpath)
        fontname = font_manager.FontProperties(fname=fontpath).get_name()  # type: ignore

        if set_font_for_matplot and fontname:
            rc("font", family=fontname)
            plt.rcParams["axes.unicode_minus"] = False
            if verbose:
                font_family = plt.rcParams["font.family"]
                logger.info(f"font family: {font_family}")
        if verbose:
            logger.info(f"font name: {fontname}")
    else:
        logger.warning(f"Font file does not exist at {fontpath}")
        fontname = ""
        fontpath = ""
        if platform.system() == "Linux":
            font_install_help = """
            apt install fontconfig
            apt install fonts-nanum
            fc-list | grep -i nanum
            """
            print(font_install_help)
    return fontname, fontpath


def find_font_file(query):
    """Find font file by query string"""
    return list(
        filter(
            lambda path: query in os.path.basename(path),
            font_manager.findSystemFonts(),
        )
    )



"""This module contains the BatchConfig class, which is used to configure batch processing jobs.

The BatchConfig class inherits from the BaseConfig class and defines a set of configuration options
that can be used to customize batch processing jobs. These options include the batch name, batch number,
batch root directory, output suffix, output extension, random seed, device, number of devices, number of
workers, and configuration file names and directories.

The BatchConfig class also defines a set of methods for setting and validating configuration options,
including set_batch_num() and _validate_batch_num() for the batch number, and _validate_seed() for the
random seed.

Example usage:

    from hyfi.batch import BatchConfig

    config = BatchConfig(batch_name="my_batch", batch_num=1, batch_root="data", output_suffix="_out",
                         output_extension=".txt", random_seed=True, seed=42, resume_run=False,
                         resume_latest=False, device="cuda", num_devices=2, num_workers=4,
                         config_yaml="config.yaml", config_json="config.json", config_dirname="configs")
"""
import random
from pathlib import Path
from typing import Optional

from pydantic import FieldValidationInfo, field_validator

from hyfi.composer.base import BaseConfig
from hyfi.utils.logging import LOGGING

logger = LOGGING.getLogger(__name__)


class BatchConfig(BaseConfig):
    """
    Configuration class for batch processing.

    Attributes:
        batch_name (str): Name of the batch.
        batch_num (Optional[int]): Number of the batch. If None, it will be set to -1.
        batch_root (str): Root directory for the batch.
        output_suffix (str): Suffix to be added to the output file name.
        output_extention (str): Extension of the output file.
        random_seed (bool): Whether to use a random seed for the batch.
        seed (int): Seed to be used for the batch. If random_seed is True or seed is None or negative, a random seed will be generated.
        resume_run (bool): Whether to resume a previous run of the batch.
        resume_latest (bool): Whether to resume the latest run of the batch.
        device (str): Device to be used for the batch.
        num_devices (int): Number of devices to be used for the batch.
        num_workers (int): Number of workers to be used for the batch.
        config_yaml (str): Name of the YAML configuration file.
        config_json (str): Name of the JSON configuration file.
        config_dirname (str): Name of the directory for the configuration files.
    """

    _config_name_: str = "__init__"
    _config_group_: str = "batch"

    batch_name: str
    batch_num: Optional[int] = None
    batch_root: str = "workspace/outputs"
    output_suffix: str = ""
    output_extention: str = ""
    random_seed: bool = True
    seed: int = -1
    resume_run: bool = False
    resume_latest: bool = False
    device: str = "cpu"
    num_devices: int = 1
    num_workers: int = 1
    config_yaml: str = "config.yaml"
    config_json: str = "config.json"
    config_dirname: str = "configs"

    _property_set_methods_ = {
        "batch_num": "set_batch_num",
    }

    def __init__(self, **data):
        """
        Initializes the BatchConfig object.

        Args:
            **data (dict): Dictionary containing the configuration data.
        """
        super().__init__(**data)
        self.set_batch_num(self.batch_num)

    def set_batch_num(self, val):
        """
        Sets the batch number.

        Args:
            val (int): Batch number.
        """
        if val is None:
            self.batch_num = -1
        if val < 0:
            num_files = len(list(self.config_dir.glob(self.config_filepattern)))
            self.batch_num = (
                num_files - 1 if self.resume_latest and num_files > 0 else num_files
            )
        if self.verbose:
            logger.info(
                "Init batch - Batch name: %s, Batch num: %s",
                self.batch_name,
                self.batch_num,
            )

    @field_validator("batch_num", mode="before")
    def _validate_batch_num(cls, v):
        """
        Validates the batch number.

        Args:
            v (int): Batch number.

        Returns:
            int: Validated batch number.
        """
        return v if v is not None else -1

    @field_validator("seed")
    def _validate_seed(cls, v, info: FieldValidationInfo):
        """
        Validates the seed.

        Args:
            v (int): Seed.
            info (FieldValidationInfo): Validation information.

        Returns:
            int: Validated seed.
        """
        if info.data["random_seed"] or v is None or v < 0:
            random.seed()
            seed = random.randint(0, 2**32 - 1)
            if info.data.get("verbose"):
                logger.info(f"Setting seed to {seed}")
            return seed
        return v

    @field_validator("output_suffix", mode="before")
    def _validate_output_suffix(cls, v):
        """
        Validates the output suffix.

        Args:
            v (str): Output suffix.

        Returns:
            str: Validated output suffix.
        """
        return v or ""

    @field_validator("output_extention", mode="before")
    def _validate_output_extention(cls, v):
        return v.strip(".") if v else ""

    @property
    def root_dir(self):
        """
        Root directory for the batch.
        """
        return Path(self.batch_root)

    @property
    def batch_dir(self):
        """
        Returns the path to the batch directory.
        """
        path_ = self.root_dir / self.batch_name
        path_.mkdir(parents=True, exist_ok=True)
        return path_

    @property
    def file_prefix(self):
        """
        Prefix for the output file name.
        """
        return f"{self.batch_name}({self.batch_num})"

    @property
    def output_file(self):
        """
        Output file name.
        """
        if self.output_suffix:
            return f"{self.file_prefix}_{self.output_suffix}.{self.output_extention}"
        else:
            return f"{self.file_prefix}.{self.output_extention}"

    @property
    def config_filename(self):
        """
        Name of the YAML configuration file.
        """
        return f"{self.file_prefix}_{self.config_yaml}"

    @property
    def config_jsonfile(self):
        """
        Name of the JSON configuration file.
        """
        return f"{self.file_prefix}_{self.config_json}"

    @property
    def config_filepattern(self):
        """
        File pattern for the configuration files.
        """
        return f"{self.batch_name}(*)_{self.config_yaml}"

    @property
    def config_dir(self):
        """
        Directory for the configuration files.
        """
        config_dir = self.batch_dir / self.config_dirname
        config_dir.mkdir(parents=True, exist_ok=True)
        return config_dir

    @property
    def config_filepath(self):
        """
        Path to the YAML configuration file.
        """
        return self.config_dir / self.config_filename

    @property
    def config_jsonpath(self):
        """
        Path to the JSON configuration file.
        """
        return self.config_dir / self.config_jsonfile



"""Collage of images.""" ""
import io
import logging
import os
import textwrap
from typing import List, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
from PIL import Image, ImageDraw
from pydantic import BaseModel, ConfigDict

from hyfi.graphics.utils import get_image_font, load_image, load_images, scale_image

log = logging.getLogger(__name__)


class Collage(BaseModel):
    """Collage of images."""

    image: Image.Image
    width: int
    height: int
    ncols: int
    nrows: int
    filepath: Optional[str] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)


def collage(
    images_or_uris,
    collage_filepath=None,
    ncols=3,
    max_images=12,
    collage_width=1200,
    padding: int = 10,
    bg_color: str = "black",
    crop_to_min_size=False,
    show_filename=False,
    filename_offset=(5, 5),
    fontname=None,
    fontsize=12,
    fontcolor="#000",
    **kwargs,
) -> Optional[Collage]:
    """
    Create a collage of images.
    """

    if not isinstance(images_or_uris, list):
        images_or_uris = [images_or_uris]
    images_or_uris = [
        uri if isinstance(uri, Image.Image) else str(uri) for uri in images_or_uris
    ]
    if not images_or_uris:
        log.info("No images provided")
        return None

    if max_images is not None:
        max_images = min(max_images, len(images_or_uris))
    else:
        max_images = len(images_or_uris)
    if max_images < 1:
        raise ValueError("max_images must be greater than 0")
    if ncols is not None and ncols > max_images:
        ncols = max_images
    if ncols is not None and max_images % ncols != 0:
        max_images = (max_images // ncols) * ncols
    # calc number of columns and rows from max_images_per_collage
    if ncols is None or ncols > max_images or ncols < 1:
        ncols = max_images // 2
    log.info(
        f"Creating collage of {max_images} images with {ncols} columns from {len(images_or_uris)} images"
    )
    img_width = collage_width // ncols
    collage_width = ncols * img_width + padding * (ncols + 1)

    # load images
    images = load_images(
        images_or_uris[:max_images],
        resize_to_multiple_of=None,
        crop_to_min_size=crop_to_min_size,
        max_width=img_width,
        **kwargs,
    )
    filenames = [
        os.path.basename(image_or_uri) if isinstance(image_or_uri, str) else None
        for image_or_uri in images_or_uris[:max_images]
    ]
    # convert images
    images = [
        convert_image(
            image,
            show_filename=show_filename,
            filename=filename,
            filename_offset=filename_offset,
            fontname=fontname,
            fontsize=fontsize,
            fontcolor=fontcolor,
        )
        for image, filename in zip(images, filenames)
    ]

    collage = grid_of_images(images, ncols, padding, bg_color=bg_color)
    if collage_filepath is not None:
        collage_filepath = str(collage_filepath)
        os.makedirs(os.path.dirname(collage_filepath), exist_ok=True)
        collage.image.save(collage_filepath)
        collage.filepath = collage_filepath
        log.info(f"Saved collage to {collage_filepath}")
    return collage


def label_collage(
    collage: Collage,
    collage_filepath=None,
    title=None,
    title_fontsize=10,
    xlabel=None,
    ylabel=None,
    xticklabels=None,
    yticklabels=None,
    xlabel_fontsize=12,
    ylabel_fontsize=12,
    dpi=100,
    fg_fontcolor="white",
    bg_color="black",
    caption=None,
    **kwargs,
) -> Collage:
    """
    Create a collage of images.
    """
    figsize = (collage.width / dpi, collage.height / dpi)
    ncols, nrows = collage.ncols, collage.nrows

    fig = plt.figure(figsize=figsize, dpi=dpi)
    fig.patch.set_facecolor(bg_color)  # type: ignore
    plt.imshow(np.array(collage.image))
    ax = plt.gca()
    plt.grid(False)
    if xlabel is None and ylabel is None:
        plt.axis("off")
    if title is not None:
        title = "\n".join(
            sum(
                (
                    textwrap.wrap(
                        t, width=int(collage.width / 15 * 12 / title_fontsize)
                    )
                    for t in title.split("\n")
                ),
                [],
            )
        )
        ax.set_title(title, fontsize=title_fontsize, color=fg_fontcolor)
    if xlabel is not None:
        # plt.xlabel(xlabel, fontdict={"fontsize": xlabel_fontsize})
        ax.set_xlabel(xlabel, fontsize=xlabel_fontsize, color=fg_fontcolor)
    if ylabel is not None:
        # plt.ylabel(ylabel, fontdict={"fontsize": ylabel_fontsize})
        ax.set_ylabel(ylabel, fontsize=ylabel_fontsize, color=fg_fontcolor)
    if xticklabels is not None:
        # get ncols number of xticks from xlim
        xlim = ax.get_xlim()
        xticks = _get_ticks(xlim, ncols)
        ax.set_xticks(xticks, color=fg_fontcolor)
        xticklabels = [""] + xticklabels
        ax.set_xticklabels(xticklabels, fontsize=xlabel_fontsize, color=fg_fontcolor)
    if yticklabels is not None:
        # get nrows number of yticks from ylim
        ylim = ax.get_ylim()
        yticks = _get_ticks(ylim, nrows)
        ax.set_yticks(yticks, color=fg_fontcolor)
        yticklabels = [""] + yticklabels
        ax.set_yticklabels(yticklabels, fontsize=ylabel_fontsize, color=fg_fontcolor)

    plt.tight_layout()
    if caption is not None:
        print(f"[{caption}]")
    img = fig2img(fig, dpi=dpi)
    img = scale_image(img, max_width=collage.width)
    plt.close()

    if collage_filepath is not None:
        collage_filepath = str(collage_filepath)
        os.makedirs(os.path.dirname(collage_filepath), exist_ok=True)
        # fig.savefig(collage_filepath, dpi=dpi, bbox_inches="tight", pad_inches=0)
        img.save(collage_filepath)
        # collage_image.save(collage_filepath)
        # log.info(f"Saved collage to {collage_filepath}")

    return Collage(
        image=img,
        filepath=collage_filepath,
        width=img.width,
        height=img.height,
        ncols=ncols,
        nrows=nrows,
    )


def _get_ticks(lim, n):
    result = np.linspace(lim[0], lim[1], n + 1)
    result = result - (result[1] - result[0]) / 2
    result[0] = lim[0]
    return result


def grid_of_images(
    images: List[Image.Image],
    ncols: int = 3,
    padding: int = 10,
    bg_color: str = "black",
) -> Collage:
    """
    Create a grid of images.
    """
    nrows = len(images) // ncols
    assert len(images) == nrows * ncols
    width, height = images[0].size
    grid_width = ncols * width + padding * (ncols + 1)
    grid_height = nrows * height + padding * (nrows + 1)
    collage = Image.new("RGB", size=(grid_width, grid_height), color=bg_color)
    for j, image in enumerate(images):
        x = j % ncols
        y = j // ncols
        collage.paste(
            image, (x * width + padding * (x + 1), y * height + padding * (y + 1))
        )
    return Collage(
        image=collage,
        width=grid_width,
        height=grid_height,
        ncols=ncols,
        nrows=nrows,
    )


def fig2img(fig, dpi=300):
    """Convert a Matplotlib figure to a PIL Image and return it"""
    buf = io.BytesIO()
    fig.savefig(buf, format="png", dpi=dpi, bbox_inches="tight", pad_inches=0)
    buf.seek(0)
    return Image.open(buf)


def convert_image(
    image_or_uri,
    show_filename: bool = False,
    filename: Optional[str] = None,
    filename_offset: Tuple[float, float] = (5, 5),
    fontname: Optional[str] = None,
    fontsize: int = 12,
    fontcolor: Optional[str] = None,
) -> Image.Image:
    """
    Convert an image to a PIL Image.
    """
    img = load_image(image_or_uri)
    if isinstance(image_or_uri, str) and filename is None:
        filename = os.path.basename(image_or_uri)
    if show_filename and filename is not None:
        font = get_image_font(fontname, fontsize)
        draw = ImageDraw.Draw(img)
        draw.text(filename_offset, filename, font=font, fill=fontcolor)

    # img = img.convert("RGB")
    # img = np.array(img)
    return img


def gallery(array: np.ndarray, ncols: int = 7):
    """
    Create a gallery of images from a numpy array.
    """
    nindex, height, width, intensity = array.shape
    nrows = nindex // ncols
    assert nindex == nrows * ncols
    return (
        array.reshape(nrows, ncols, height, width, intensity)
        .swapaxes(1, 2)
        .reshape(height * nrows, width * ncols, intensity)
    )



"""This module contains the About Configuration for the HyFI package.

It defines the AboutConfig class, which is a Pydantic BaseModel that contains
metadata about the package, such as its name, version, authors, and license.
It also defines the model_config attribute, which is a ConfigDict that allows
extra configuration options to be added to the AboutConfig instance.
"""
from pydantic import BaseModel, ConfigDict

from hyfi.utils.logging import LOGGING

logger = LOGGING.getLogger(__name__)

__hyfi_name__: str = "HyFI"
__hyfi_authors__: str = "Young Joon Lee <entelecheia@hotmail.com>"
__hyfi_description__: str = (
    "Hydra Fast Interface (Hydra and Pydantic based interface framework)"
)
__hyfi_homepage__: str = "https://hyfi.entelecheia.ai"
__hyfi_license__: str = "MIT"


class AboutConfig(BaseModel):
    """A Pydantic BaseModel that contains metadata about the package.

    Attributes:
        name (str): The display name of the package.
        authors (str): The author(s) of the package.
        description (str): A brief description of the package.
        homepage (str): The URL of the package's homepage.
        license (str): The license under which the package is distributed.
    """

    _config_group_: str = "about"

    name: str = __hyfi_name__
    authors: str = __hyfi_authors__
    description: str = __hyfi_description__
    homepage: str = __hyfi_homepage__
    license: str = __hyfi_license__

    model_config = ConfigDict(extra="allow")  # type: ignore



from typing import Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, model_validator

from hyfi.composer import Composer
from hyfi.project import ProjectConfig
from hyfi.task import TaskConfig
from hyfi.utils.logging import LOGGING

logger = LOGGING.getLogger(__name__)

Tasks = List[TaskConfig]


class WorkflowConfig(BaseModel):
    project: Optional[ProjectConfig] = None
    tasks: Optional[List[Union[str, Dict]]] = []
    verbose: bool = False

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="allow",
        validate_assignment=False,
    )  # type: ignore

    @model_validator(mode="before")
    def validate_model_config_before(cls, data):
        # logger.debug("Validating model config before validating each field.")
        return Composer.to_dict(data)

    def get_tasks(self) -> Tasks:
        self.tasks = self.tasks or []
        tasks: Tasks = []
        for name in self.tasks:
            if isinstance(name, str) and isinstance(getattr(self, name), dict):
                task = TaskConfig(**getattr(self, name))
                task.name = name
                tasks.append(task)
        return tasks
